package com.linbit.linstor.core.apicallhandler.response;

import com.linbit.linstor.api.ApiConsts;

import java.util.UUID;

public class ApiOperation
{
    private final long opMask;

    private final OperationDescription description;

    public ApiOperation(long opMaskRef, OperationDescription descriptionRef)
    {
        opMask = opMaskRef;
        description = descriptionRef;
    }

    public long getOpMask()
    {
        return opMask;
    }

    public OperationDescription getDescription()
    {
        return description;
    }

    /**
     * The term 'create' is used for Linstor internal constructs without any corresponding external entity.
     * Corresponds to success messages generated by {@link ApiSuccessUtils#defaultCreatedEntry(UUID, String)}.
     */
    public static ApiOperation makeCreateOperation()
    {
        return new ApiOperation(ApiConsts.MASK_CRT, new OperationDescription("creation", "creating"));
    }

    /**
     * The term 'register' is used for entities with a corresponding external entity.
     * Corresponds to success messages generated by {@link ApiSuccessUtils#defaultRegisteredEntry(UUID, String)}.
     */
    public static ApiOperation makeRegisterOperation()
    {
        return new ApiOperation(ApiConsts.MASK_CRT, new OperationDescription("registration", "registering"));
    }

    public static ApiOperation makeModifyOperation()
    {
        return new ApiOperation(ApiConsts.MASK_MOD, new OperationDescription("modification", "modifying"));
    }

    public static ApiOperation makeDeleteOperation()
    {
        return new ApiOperation(ApiConsts.MASK_DEL, new OperationDescription("deletion", "deleting"));
    }
}
